#!/usr/bin/python
"""variety of (non-display) image utilities
"""
import numpy as np
from ..pyramids.filters import named_filter
from ..pyramids.c.wrapper import corrDn, upConv


def _init_filt(filt):
    if isinstance(filt, str):
        filt = named_filter(filt)
    else:
        filt = np.array(filt)
    return filt / filt.sum()


def blur(image, n_levels=1, filt='binom5'):
    '''blur an image by filtering and downsampling then by upsampling and filtering

    Blur an image, by filtering and downsampling N_LEVELS times (default=1), followed by upsampling
    and filtering LEVELS times.  The blurring is done with filter kernel specified by FILT (default
    = 'binom5'), which can be a string (to be passed to named_filter), a vector (applied separably
    as a 1D convolution kernel in X and Y), or a matrix (applied as a 2D convolution kernel).  The
    downsampling is always by 2 in each direction.

    This differs from blurDn in that here we upsample afterwards.

    Arguments
    ---------
    image : `array_like`
        1d or 2d image to blur
    n_levels : `int`
        the number of times to filter and downsample. the higher this is, the more blurred the
        resulting image will be
    filt : {`array_like`, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
            'qmf5', 'qmf9', 'qmf13'}
        filter to use for filtering image. If array_like, can be 1d or 2d. All scaled so L-1 norm
        is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_

    Returns
    -------
    image : `array_like`
        the blurred image

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.
    '''

    filt = _init_filt(filt)

    if n_levels > 0:
        if len(image.shape) == 1 or image.shape[0] == 1 or image.shape[1] == 1:
            # 1D image
            if len(filt) == 2 and (np.asarray(filt.shape) != 1).any():
                raise Exception('Error: can not apply 2D filterer to 1D signal')

            imIn = corrDn(image, filt, 'reflect1', len(image))
            out = blur(imIn, n_levels-1, filt)
            res = upConv(out, filt, 'reflect1', len(image), [0, 0], len(image))
            return res
        elif len(filt.shape) == 1 or filt.shape[0] == 1 or filt.shape[1] == 1:
            # 2D image 1D filter
            imIn = corrDn(image, filt, 'reflect1', [2, 1])
            imIn = corrDn(imIn, filt.T, 'reflect1', [1, 2])
            out = blur(imIn, n_levels-1, filt)
            res = upConv(out, filt.T, 'reflect1', [1, 2], [0, 0],
                         [out.shape[0], image.shape[1]])
            res = upConv(res, filt, 'reflect1', [2, 1], [0, 0],
                         image.shape)
            return res
        else:
            # 2D image 2D filter
            imIn = corrDn(image, filt, 'reflect1', [2, 2])
            out = blur(imIn, n_levels-1, filt)
            res = upConv(out, filt, 'reflect1', [2, 2], [0, 0],
                         image.shape)
            return res
    else:
        return image


def blurDn(image, n_levels=1, filt='binom5'):
    '''blur and downsample an image

    Blur and downsample an image.  The blurring is done with filter kernel specified by FILT
    (default = 'binom5'), which can be a string (to be passed to named_filter), a vector (applied
    separably as a 1D convolution kernel in X and Y), or a matrix (applied as a 2D convolution
    kernel).  The downsampling is always by 2 in each direction.

    The procedure is applied recursively `n_levels` times (default=1).

    This differs from blur in that we do NOT upsample afterwards.

    Arguments
    ---------
    image : `array_like`
        1d or 2d image to blur and downsample
    n_levels : `int`
        the number of times to filter and downsample. the higher this is, the blurrier and smaller
        the resulting image will be
    filt : {`array_like`, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
            'qmf5', 'qmf9', 'qmf13'}
        filter to use for filtering image. If array_like, can be 1d or 2d. All scaled so L-1 norm
        is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_

    Returns
    -------
    image : `array_like`
        the blurred and downsampled image

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.
    '''

    filt = _init_filt(filt)

    if n_levels > 1:
        image = blurDn(image, n_levels-1, filt)

    if n_levels >= 1:
        if len(image.shape) == 1 or image.shape[0] == 1 or image.shape[1] == 1:
            # 1D image
            if len(filt.shape) > 1 and (filt.shape[1] != 1 and filt.shape[2] != 1):
                # >1D filter
                raise Exception('Error: Cannot apply 2D filter to 1D signal')

            # orient filter and image correctly
            if image.shape[0] == 1:
                if len(filt.shape) == 1 or filt.shape[1] == 1:
                    filt = filt.T
            else:
                if filt.shape[0] == 1:
                    filt = filt.T

            res = corrDn(image=image, filt=filt, step=(2, 2))
            if len(image.shape) == 1 or image.shape[1] == 1:
                res = np.reshape(res, (np.ceil(image.shape[0]/2.0).astype(int), 1))
            else:
                res = np.reshape(res, (1, np.ceil(image.shape[1]/2.0).astype(int)))
        elif len(filt.shape) == 1 or filt.shape[0] == 1 or filt.shape[1] == 1:
            # 2D image and 1D filter
            res = corrDn(image=image, filt=filt.T, step=(2, 1))
            res = corrDn(image=res, filt=filt, step=(1, 2))

        else:  # 2D image and 2D filterer
            res = corrDn(image=image, filt=filt, step=(2, 2))
    else:
        res = image

    return res


def upBlur(image, n_levels=1, filt='binom5'):
    '''upsample and blur an image.

    Upsample and blur an image.  The blurring is done with filter kernel specified by FILT (default
    = 'binom5'), which can be a string (to be passed to named_filter), a vector (applied separably
    as a 1D convolution kernel in X and Y), or a matrix (applied as a 2D convolution kernel).  The
    downsampling is always by 2 in each direction.

    The procedure is applied recursively n_levels times (default=1).

    Arguments
    ---------
    image : `array_like`
        1d or 2d image to upsample and blur
    n_levels : `int`
        the number of times to filter and downsample. the higher this is, the blurrier and larger
        the resulting image will be
    filt : {`array_like`, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
            'qmf5', 'qmf9', 'qmf13'}
        filter to use for filtering image. If array_like, can be 1d or 2d. All scaled so L-1 norm
        is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_

    Returns
    -------
    image : `array_like`
        the upsampled and blurred image

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.

    '''

    if isinstance(filt, str):
        filt = named_filter(filt)
    # print(filt, n_levels)

    if n_levels > 1:
        image = upBlur(image, n_levels-1, filt)

    if n_levels >= 1:
        # 1d image
        if image.shape[0] == 1 or image.shape[1] == 1:
            if image.shape[0] == 1:
                filt = filt.reshape(filt.shape[1], filt.shape[0])
                start = (1, 2)
            else:
                start = (2, 1)
            res = upConv(image, filt, 'reflect1', start)
        elif filt.shape[0] == 1 or filt.shape[1] == 1:
            if filt.shape[0] == 1:
                filt = filt.reshape(filt.shape[1], 1)
            res = upConv(image, filt, 'reflect1', [2, 1])
            res = upConv(res, filt.T, 'reflect1', [1, 2])
        else:
            res = upConv(image, filt, 'reflect1', [2, 2])
    else:
        res = image

    return res


def image_gradient(image, edge_type="dont-compute"):
    '''Compute the gradient of the image using smooth derivative filters

    Compute the gradient of the image using smooth derivative filters optimized for accurate
    direction estimation.  Coordinate system corresponds to standard pixel indexing: X axis points
    rightward.  Y axis points downward.  `edges` specify boundary handling.

    Notes
    -----
    original filters from Int'l Conf Image Processing, 1994.
    updated filters 10/2003: see Farid & Simoncelli, IEEE Trans Image
                             Processing, 13(4):496-508, April 2004.

    Arguments
    ---------
    image : `array_like`
        2d array to compute the gradients of
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Returns
    -------
    dx, dy : `np.array`
        the X derivative and the Y derivative

    '''

    # kernels from Farid & Simoncelli, IEEE Trans Image Processing,
    #   13(4):496-508, April 2004.
    gp = np.array([0.037659,  0.249153, 0.426375, 0.249153, 0.037659]).reshape(5, 1)
    gd = np.array([-0.109604, -0.276691, 0.000000, 0.276691, 0.109604]).reshape(5, 1)

    dx = corrDn(corrDn(image, gp, edge_type), gd.T, edge_type)
    dy = corrDn(corrDn(image, gd, edge_type), gp.T, edge_type)

    return (dx, dy)



import numpy as np
import warnings


def matlab_round(array):
    '''round equivalent to matlab function, which rounds .5 away from zero

    (used in matlab_histo so we can unit test against matlab code).

    But np.round() would rounds .5 to nearest even number, e.g.:
    - np.round(0.5) = 0, matlab_round(0.5) = 1
    - np.round(2.5) = 2, matlab_round(2.5) = 3

    Arguments
    ---------
    array : `np.array`
        the array to round

    Returns
    -------
    rounded_array : `np.array`
        the rounded array
    '''
    (fracPart, intPart) = np.modf(array)
    return intPart + (np.abs(fracPart) >= 0.5) * np.sign(fracPart)


def matlab_histo(array, nbins=101, binsize=None, center=None):
    '''Compute a histogram of array.

    How does this differ from MatLab's HIST function?  This function:
      - allows uniformly spaced bins only.
      + operates on all elements of MTX, instead of columnwise.
      + is much faster (approximately a factor of 80 on my machine).
      + allows specification of number of bins OR binsize. Default=101 bins.
      + allows (optional) specification of binCenter.

    Arguments
    ---------
    array : `np.array`
        the array to bin
    nbins : `int`
        the number of histogram bins
    binsize : `float` or None
        the size of each bin. if None, we use nbins to determine it as:
        `(array.max() - array.min()) / nbins`
    center : `float` or None
        the center position for the histogram bins. if None, this is `array.mean()`

    Returns
    -------
    N : `np.array`
        the histogram counts
    edges : `np.array`
        vector containing the centers of the histogram bins
    '''
    mini = array.min()
    maxi = array.max()

    if center is None:
        center = array.mean()

    if binsize is None:
        # use nbins to determine binsize
        binsize = (maxi-mini) / nbins

    nbins2 = int(matlab_round((maxi - center) / binsize) - matlab_round((mini - center) / binsize))
    if nbins2 != nbins:
        warnings.warn('Overriding bin number %d (requested %d)' % (nbins2, nbins))
        nbins = nbins2

    # np.histogram uses bin edges, not centers like Matlab's hist
    # compute bin edges (nbins + 1 of them)
    edge_left = center + binsize * (-0.499 + matlab_round((mini - center) / binsize))
    edges = edge_left + binsize * np.arange(nbins+1)
    N, _ = np.histogram(array, edges)

    # matlab version returns column vectors, so we will too.
    # to check: return edges or centers? edit comments.
    return (N.reshape(1, -1), edges.reshape(1, -1))


def rcosFn(width=1, position=0, values=(0, 1)):
    '''Return a lookup table containing a "raised cosine" soft threshold function

    Y =  VALUES(1) + (VALUES(2)-VALUES(1)) * cos^2( PI/2 * (X - POSITION + WIDTH)/WIDTH )

    this lookup table is suitable for use by `pointOp`

    Arguments
    ---------
    width : `float`
        the width of the region over which the transition occurs
    position : `float`
        the location of the center of the threshold
    values : `tuple`
        2-tuple specifying the values to the left and right of the transition.

    Returns
    -------
    X : `np.array`
        the x valuesof this raised cosine
    Y : `np.array`
        the y valuesof this raised cosine
    '''

    sz = 256   # arbitrary!

    X = np.pi * np.arange(-sz-1, 2) / (2*sz)

    Y = values[0] + (values[1]-values[0]) * np.cos(X)**2

    # make sure end values are repeated, for extrapolation...
    Y[0] = Y[1]
    Y[sz+2] = Y[sz+1]

    X = position + (2*width/np.pi) * (X + np.pi/4)

    return (X, Y)


if __name__ == "__main__":
    X, Y = rcosFn(width=1, position=0, values=(0, 1))

    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(X, Y)
    plt.show()



from . import pyramids

from .pyramids.c.wrapper import corrDn, upConv, pointOp

from .pyramids.filters import named_filter, binomial_filter, steerable_filters

from .tools.display_tools import imshow, animshow, pyrshow
from .tools import synthetic_images
from .tools.image_stats import image_compare, image_stats, range, skew, var, entropy
from .tools.utils import rcosFn, matlab_histo, matlab_round
from .tools.convolutions import rconv2
from .tools.comparePyr import comparePyr
from .tools.compareRecon import compareRecon

from .tools.extra_tools import blurDn, blur, upBlur, image_gradient

from .version import version as __version__



import numpy as np
import scipy.signal

# ----------------------------------------------------------------
# Below are (slow) scipy convolution functions
# they are intended for comparison purpose only
# the c code is prefered and used throughout this package
# ----------------------------------------------------------------


def rconv2(mtx1, mtx2, ctr=0):
    '''Convolution of two matrices, with boundaries handled via reflection about the edge pixels.

    Result will be of size of LARGER matrix.

    The origin of the smaller matrix is assumed to be its center.
    For even dimensions, the origin is determined by the CTR (optional)
    argument:
         CTR   origin
          0     DIM/2      (default)
          1   (DIM/2)+1

    In general, you should not use this function, since it will be slow. Instead, use `upConv` or
    `corrDn`, which use the C code and so are much faster.

    '''

    if (mtx1.shape[0] >= mtx2.shape[0] and mtx1.shape[1] >= mtx2.shape[1]):
        large = mtx1
        small = mtx2
    elif (mtx1.shape[0] <= mtx2.shape[0] and mtx1.shape[1] <= mtx2.shape[1]):
        large = mtx2
        small = mtx1
    else:
        print('one matrix must be larger than the other in both dimensions!')
        return

    ly = large.shape[0]
    lx = large.shape[1]
    sy = small.shape[0]
    sx = small.shape[1]

    # These values are one less than the index of the small mtx that falls on
    # the border pixel of the large matrix when computing the first
    # convolution response sample:
    sy2 = int(np.floor((sy+ctr-1)/2))
    sx2 = int(np.floor((sx+ctr-1)/2))

    # pad with reflected copies
    nw = large[sy-sy2-1:0:-1, sx-sx2-1:0:-1]
    n = large[sy-sy2-1:0:-1, :]
    ne = large[sy-sy2-1:0:-1, lx-2:lx-sx2-2:-1]
    w = large[:, sx-sx2-1:0:-1]
    e = large[:, lx-2:lx-sx2-2:-1]
    sw = large[ly-2:ly-sy2-2:-1, sx-sx2-1:0:-1]
    s = large[ly-2:ly-sy2-2:-1, :]
    se = large[ly-2:ly-sy2-2:-1, lx-2:lx-sx2-2:-1]

    n = np.column_stack((nw, n, ne))
    c = np.column_stack((w, large, e))
    s = np.column_stack((sw, s, se))

    clarge = np.concatenate((n, c), axis=0)
    clarge = np.concatenate((clarge, s), axis=0)

    return scipy.signal.convolve(clarge, small, 'valid')


# def cconv2(mtx1, mtx2, ctr=0):
#     '''Circular convolution of two matrices.  Result will be of size of
#     LARGER vector.
#
#     The origin of the smaller matrix is assumed to be its center.
#     For even dimensions, the origin is determined by the CTR (optional)
#     argument:
#          CTR   origin
#           0     DIM/2      (default)
#           1     (DIM/2)+1
#
#     Eero Simoncelli, 6/96.  Modified 2/97.
#     Python port by Rob Young, 8/15
#     '''
#
#     if len(args) < 2:
#         print 'Error: cconv2 requires two input matrices!'
#         print 'Usage: cconv2(matrix1, matrix2, center)'
#         print 'where center parameter is optional'
#         return
#     else:
#         a = np.array(args[0])
#         b = np.array(args[1])
#
#     if len(args) == 3:
#         ctr = args[2]
#     else:
#         ctr = 0
#
#     if a.shape[0] >= b.shape[0] and a.shape[1] >= b.shape[1]:
#         large = a
#         small = b
#     elif a.shape[0] <= b.shape[0] and a.shape[1] <= b.shape[1]:
#         large = b
#         small = a
#     else:
#         print 'Error: one matrix must be larger than the other in both dimensions!'
#         return
#
#     ly = large.shape[0]
#     lx = large.shape[1]
#     sy = small.shape[0]
#     sx = small.shape[1]
#
#     ## These values are the index of the small mtx that falls on the
#     ## border pixel of the large matrix when computing the first
#     ## convolution response sample:
#     sy2 = np.floor((sy+ctr+1)/2.0).astype(int)
#     sx2 = np.floor((sx+ctr+1)/2.0).astype(int)
#
#     # pad
#     nw = large[ly-sy+sy2:ly, lx-sx+sx2:lx]
#     n = large[ly-sy+sy2:ly, :]
#     ne = large[ly-sy+sy2:ly, :sx2-1]
#     w = large[:, lx-sx+sx2:lx]
#     c = large
#     e = large[:, :sx2-1]
#     sw = large[:sy2-1, lx-sx+sx2:lx]
#     s = large[:sy2-1, :]
#     se = large[:sy2-1, :sx2-1]
#
#     n = np.column_stack((nw, n, ne))
#     c = np.column_stack((w,large,e))
#     s = np.column_stack((sw, s, se))
#
#     clarge = np.concatenate((n, c), axis=0)
#     clarge = np.concatenate((clarge, s), axis=0)
#
#     c = scipy.signal.convolve(clarge, small, 'valid')
#
#     return c


# def zconv2(mtx1, mtx2, ctr=0):
#     ''' RES = ZCONV2(MTX1, MTX2, CTR)
#
#         Convolution of two matrices, with boundaries handled as if the larger
#         mtx lies in a sea of zeros. Result will be of size of LARGER vector.
#
#         The origin of the smaller matrix is assumed to be its center.
#         For even dimensions, the origin is determined by the CTR (optional)
#         argument:
#              CTR   origin
#               0     DIM/2      (default)
#               1     (DIM/2)+1  (behaves like conv2(mtx1,mtx2,'same'))
#
#         Eero Simoncelli, 2/97.  Python port by Rob Young, 10/15.  '''
#
#     # REQUIRED ARGUMENTS
#     #----------------------------------------------------------------
#
#     if len(args) < 2 or len(args) > 3:
#         print 'Usage: zconv2(matrix1, matrix2, center)'
#         print 'first two input parameters are required'
#         return
#     else:
#         a = np.array(args[0])
#         b = np.array(args[1])
#
#     # OPTIONAL ARGUMENT
#     #----------------------------------------------------------------
#
#     if len(args) == 3:
#         ctr = args[2]
#     else:
#         ctr = 0
#
#     #----------------------------------------------------------------
#
#     if (a.shape[0] >= b.shape[0]) and (a.shape[1] >= b.shape[1]):
#         large = a
#         small = b
#     elif (a.shape[0] <= b.shape[0]) and (a.shape[1] <= b.shape[1]):
#         large = b
#         small = a
#     else:
#         print 'Error: one arg must be larger than the other in both dimensions!'
#         return
#
#     ly = large.shape[0]
#     lx = large.shape[1]
#     sy = small.shape[0]
#     sx = small.shape[1]
#
#     ## These values are the index of the small matrix that falls on the
#     ## border pixel of the large matrix when computing the first
#     ## convolution response sample:
#     sy2 = np.floor((sy+ctr+1)/2.0).astype(int)-1
#     sx2 = np.floor((sx+ctr+1)/2.0).astype(int)-1
#
#     clarge = scipy.signal.convolve(large, small, 'full')
#
#     c = clarge[sy2:ly+sy2, sx2:lx+sx2]
#
#     return c



import numpy
import math


def compareRecon(recon1, recon2, rtol=1e-5, atol=1e-8):
    '''compare two arrays

    returns True is they are the same within specified precision and False if not.  function was
    made to accompany unit test code.

    This function is deprecated. Instead use the builtin numpy:

    np.allclose(recon1, recon2, rtol=1e-05, atol=1e-08, equal_nan=False)

    This will not tell you where the error is, but you can find that yourself

    '''

    # NOTE builtin numpy:

    # BUT won't print where and what the first error is

    if recon1.shape != recon2.shape:
        print('shape is different!')
        print(recon1.shape)
        print(recon2.shape)
        return False

    prec = -1
    for i in range(recon1.shape[0]):
        for j in range(recon2.shape[1]):
            if numpy.absolute(recon1[i, j].real - recon2[i, j].real) > math.pow(10, -11):
                print("real: i=%d j=%d %.15f %.15f diff=%.15f" %
                      (i, j, recon1[i, j].real, recon2[i, j].real,
                       numpy.absolute(recon1[i, j].real-recon2[i, j].real)))
                return False
            # FIX: need a better way to test
            # if we have many significant digits to the left of decimal we
            # need to be less stringent about digits to the right.
            # The code below works, but there must be a better way.
            if isinstance(recon1, complex):
                if int(math.log(numpy.abs(recon1[i, j].imag), 10)) > 1:
                    prec = prec + int(math.log(numpy.abs(recon1[i, j].imag), 10))
                    if prec > 0:
                        prec = -1
                print(prec)
                if numpy.absolute(recon1[i, j].imag - recon2[i, j].imag) > math.pow(10, prec):
                    print("imag: i=%d j=%d %.15f %.15f diff=%.15f" %
                          (i, j, recon1[i, j].imag, recon2[i, j].imag,
                           numpy.absolute(recon1[i, j].imag-recon2[i, j].imag)))
                    return False

    return True



import numpy as np
from ..pyramids import convert_pyr_coeffs_to_pyr


def comparePyr(matPyr, pyPyr, rtol=1e-5, atol=1e-8):
    '''compare two pyramids

    returns True if they are the same with in desired precision and False if not.

    written for unit testing code.

    '''
    # compare two pyramids - return 0 for !=, 1 for ==
    # correct number of elements?
    matSz = sum(matPyr.shape)
    try:
        pySz = 1 + sum([np.array(size).prod() for size in pyPyr.pyr_size.values()])
    except AttributeError:
        pySz = 1 + sum([np.array(size).prod() for size in pyPyr.pyrSize])

    if(matSz != pySz):
        print("size difference: %d != %d, returning False" % (matSz, pySz))
        return False

    # values are close to each other?
    matStart = 0
    try:
        pyCoeffs, pyHigh, pyLow = convert_pyr_coeffs_to_pyr(pyPyr.pyr_coeffs)
        if pyHigh is not None:
            pyCoeffs.insert(0, pyHigh)
        if pyLow is not None:
            pyCoeffs.append(pyLow)
    except AttributeError:
        pyCoeffs = pyPyr.pyr
    for idx, pyTmp in enumerate(pyCoeffs):
        matTmp = matPyr[matStart:matStart + pyTmp.size]
        matStart = matStart + pyTmp.size
        matTmp = np.reshape(matTmp, pyTmp.shape, order='F')

        # relative tolerance rtol
        # absolute tolerance atol
        isclose = np.isclose(matTmp, pyTmp, rtol, atol)
        if not isclose.all():
            print("some pyramid elements not identical: checking...")
            for i in range(isclose.shape[0]):
                for j in range(isclose.shape[1]):
                    if not isclose[i, j]:
                        print("failed level:%d element:%d %d value:%.15f %.15f" %
                              (idx, i, j, matTmp[i, j], pyTmp[i, j]))
                        return False

    return True
